import logging
import re
from typing import Any, Sequence

import unique_sdk
from openai.types.chat import ChatCompletionToolChoiceOptionParam
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from pydantic import BaseModel
from typing_extensions import deprecated
from unique_sdk._list_object import ListObject

from unique_toolkit._common import _time_utils
from unique_toolkit._common.token import count_tokens
from unique_toolkit.chat.constants import (
    DEFAULT_MAX_MESSAGES,
)
from unique_toolkit.chat.schemas import (
    ChatMessage,
    ChatMessageAssessment,
    ChatMessageAssessmentLabel,
    ChatMessageAssessmentStatus,
    ChatMessageAssessmentType,
    ChatMessageRole,
    MessageExecution,
    MessageExecutionType,
    MessageExecutionUpdateStatus,
    MessageLog,
    MessageLogDetails,
    MessageLogStatus,
    MessageLogUncitedReferences,
)
from unique_toolkit.content.schemas import ContentChunk, ContentReference
from unique_toolkit.language_model.constants import (
    DEFAULT_COMPLETE_TEMPERATURE,
    DEFAULT_COMPLETE_TIMEOUT,
)
from unique_toolkit.language_model.functions import (
    _prepare_all_completions_params_util,
)
from unique_toolkit.language_model.infos import (
    LanguageModelInfo,
    LanguageModelName,
)
from unique_toolkit.language_model.schemas import (
    LanguageModelMessages,
    LanguageModelStreamResponse,
    LanguageModelTool,
    LanguageModelToolDescription,
)

logger = logging.getLogger(__name__)


def modify_message(
    user_id: str,
    company_id: str,
    assistant_message_id: str,
    chat_id: str,
    user_message_id: str,
    user_message_text: str,
    assistant: bool,
    content: str | None = None,
    original_content: str | None = None,
    references: list[ContentReference] | None = None,
    debug_info: dict | None = None,
    message_id: str | None = None,
    set_completed_at: bool = False,
) -> ChatMessage:
    """Modifies a chat message synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        assistant_message_id (str): The assistant message ID.
        chat_id (str): The chat ID.
        user_message_id (str): The user message ID.
        user_message_text (str): The user message text.
        assistant (bool): Whether the message is an assistant message.
        content (str, optional): The new content for the message.
        original_content (str, optional): The original content for the message.
        message_id (str, optional): The message ID. Defaults to None, then the ChatState assistant message id is used.
        references (list[ContentReference]): list of ContentReference objects. Defaults to None.
        debug_info (dict[str, Any]], optional): Debug information. Defaults to None.
        set_completed_at (bool, optional): Whether to set the completedAt field with the current date time. Defaults to False.

    Returns:
        ChatMessage: The modified message.

    Raises:
        Exception: If the modification fails.

    """
    try:
        params = _construct_message_modify_params(
            user_id=user_id,
            company_id=company_id,
            assistant_message_id=assistant_message_id,
            chat_id=chat_id,
            user_message_id=user_message_id,
            user_message_text=user_message_text,
            assistant=assistant,
            content=content,
            original_content=original_content,
            references=references,
            debug_info=debug_info,
            message_id=message_id,
            set_completed_at=set_completed_at,
        )
        message = unique_sdk.Message.modify(**params)
        return ChatMessage(**message)
    except Exception as e:
        logger.error(f"Failed to modify user message: {e}")
        raise e


async def modify_message_async(
    user_id: str,
    company_id: str,
    assistant_message_id: str,
    chat_id: str,
    user_message_id: str,
    user_message_text: str,
    assistant: bool,
    content: str | None = None,
    original_content: str | None = None,
    references: list[ContentReference] | None = None,
    debug_info: dict | None = None,
    message_id: str | None = None,
    set_completed_at: bool = False,
) -> ChatMessage:
    """Modifies a chat message asynchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        assistant_message_id (str): The assistant message ID.
        chat_id (str): The chat ID.
        user_message_id (str): The user message ID.
        user_message_text (str): The user message text.
        assistant (bool): Whether the message is an assistant message.
        content (str, optional): The new content for the message.
        original_content (str, optional): The original content for the message.
        message_id (str, optional): The message ID. Defaults to None, then the ChatState assistant message id is used.
        references (list[ContentReference]): list of ContentReference objects. Defaults to None.
        debug_info (dict[str, Any]], optional): Debug information. Defaults to None.
        set_completed_at (bool, optional): Whether to set the completedAt field with the current date time. Defaults to False.

    Returns:
        ChatMessage: The modified message.

    Raises:
        Exception: If the modification fails.

    """
    try:
        params = _construct_message_modify_params(
            user_id=user_id,
            company_id=company_id,
            assistant_message_id=assistant_message_id,
            chat_id=chat_id,
            user_message_id=user_message_id,
            user_message_text=user_message_text,
            assistant=assistant,
            content=content,
            original_content=original_content,
            references=references,
            debug_info=debug_info,
            message_id=message_id,
            set_completed_at=set_completed_at,
        )
        message = await unique_sdk.Message.modify_async(**params)
        return ChatMessage(**message)
    except Exception as e:
        logger.error(f"Failed to modify user message: {e}")
        raise e


def map_references(references: list[ContentReference]) -> list[dict[str, Any]]:
    return [
        {
            "name": ref.name,
            "url": ref.url,
            "sequenceNumber": ref.sequence_number,
            "sourceId": ref.source_id,
            "source": ref.source,
        }
        for ref in references
    ]


def map_references_with_original_index(
    references: list[ContentReference],
) -> list[dict[str, Any]]:
    """Maps ContentReference objects to dict format for MessageLog API (complete format).

    This function includes all fields including originalIndex for APIs that support it.
    """
    return [
        {
            "name": ref.name,
            "url": ref.url,
            "sequenceNumber": ref.sequence_number,
            "sourceId": ref.source_id,
            "source": ref.source,
            "originalIndex": ref.original_index,
        }
        for ref in references
    ]


def _construct_message_modify_params(
    user_id: str,
    company_id: str,
    assistant_message_id: str,
    chat_id: str,
    user_message_id: str,
    user_message_text: str,
    assistant: bool = True,
    content: str | None = None,
    original_content: str | None = None,
    references: list[ContentReference] | None = None,
    debug_info: dict | None = None,
    message_id: str | None = None,
    set_completed_at: bool = False,
) -> dict[str, Any]:
    completed_at_datetime = None

    if message_id:
        # Message ID specified. No need to guess
        message_id = message_id
    elif assistant:
        # Assistant message ID
        message_id = assistant_message_id
    else:
        message_id = user_message_id
        if content is None:
            content = user_message_text

    if set_completed_at:
        completed_at_datetime = _time_utils.get_datetime_now()

    params = {
        "user_id": user_id,
        "company_id": company_id,
        "id": message_id,
        "chatId": chat_id,
        "text": content,
        "originalText": original_content,
        "references": map_references(references) if references else [],
        "debugInfo": debug_info,
        "completedAt": completed_at_datetime,
    }
    return params


def create_message(
    user_id: str,
    company_id: str,
    chat_id: str,
    assistant_id: str,
    role: ChatMessageRole,
    content: str | None = None,
    original_content: str | None = None,
    references: list[ContentReference] | None = None,
    debug_info: dict | None = None,
    set_completed_at: bool | None = False,
) -> ChatMessage:
    """Creates a message in the chat session synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        chat_id (str): The chat ID.
        assistant_id (str): The assistant ID.
        role (ChatMessageRole): The role of the message.
        content (str, optional): The content for the message. Defaults to None.
        original_content (str, optional): The original content for the message. Defaults to None.
        references (list[ContentReference], optional): list of ContentReference objects. Defaults to None.
        debug_info (dict[str, Any]], optional): Debug information. Defaults to None.
        set_completed_at (Optional[bool]): Whether to set the completedAt field with the current date time. Defaults to False.

    Returns:
        ChatMessage: The created message.

    Raises:
        Exception: If the creation fails.

    """
    if original_content is None:
        original_content = content

    try:
        params = _construct_message_create_params(
            user_id=user_id,
            company_id=company_id,
            chat_id=chat_id,
            assistant_id=assistant_id,
            role=role,
            content=content,
            original_content=original_content,
            references=references,
            debug_info=debug_info,
            set_completed_at=set_completed_at,
        )

        message = unique_sdk.Message.create(**params)
        return ChatMessage(**message)
    except Exception as e:
        logger.error(f"Failed to create assistant message: {e}")
        raise e


async def create_message_async(
    user_id: str,
    company_id: str,
    chat_id: str,
    assistant_id: str,
    role: ChatMessageRole,
    content: str | None = None,
    original_content: str | None = None,
    references: list[ContentReference] | None = None,
    debug_info: dict | None = None,
    set_completed_at: bool | None = False,
):
    """Creates a message in the chat session synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        chat_id (str): The chat ID.
        assistant_id (str): The assistant ID.
        role (ChatMessageRole): The role of the message.
        content (str, optional): The content for the message. Defaults to None.
        original_content (str, optional): The original content for the message. Defaults to None.
        references (list[ContentReference], optional): list of ContentReference objects. Defaults to None.
        debug_info (dict[str, Any]], optional): Debug information. Defaults to None.
        set_completed_at (Optional[bool]): Whether to set the completedAt field with the current date time. Defaults to False.

    Returns:
        ChatMessage: The created message.

    Raises:
        Exception: If the creation fails.

    """
    if original_content is None:
        original_content = content

    try:
        params = _construct_message_create_params(
            user_id=user_id,
            company_id=company_id,
            chat_id=chat_id,
            assistant_id=assistant_id,
            role=role,
            content=content,
            original_content=original_content,
            references=references,
            debug_info=debug_info,
            set_completed_at=set_completed_at,
        )

        message = await unique_sdk.Message.create_async(**params)
        return ChatMessage(**message)
    except Exception as e:
        logger.error(f"Failed to create assistant message: {e}")
        raise e


def _construct_message_create_params(
    user_id: str,
    company_id: str,
    chat_id: str,
    assistant_id: str,
    role: ChatMessageRole,
    content: str | None = None,
    original_content: str | None = None,
    references: list[ContentReference] | None = None,
    debug_info: dict | None = None,
    set_completed_at: bool | None = False,
) -> dict[str, Any]:
    if original_content is None:
        original_content = content

    return {
        "user_id": user_id,
        "company_id": company_id,
        "assistantId": assistant_id,
        "role": role.value.upper(),
        "chatId": chat_id,
        "text": content,
        "originalText": original_content,
        "references": map_references(references) if references else [],
        "debugInfo": debug_info or {},
        "completedAt": _time_utils.get_datetime_now() if set_completed_at else None,
    }


def get_selection_from_history(
    full_history: list[ChatMessage],
    max_tokens: int,
    max_messages=DEFAULT_MAX_MESSAGES,
    model_info: LanguageModelInfo | None = None,
) -> list[ChatMessage]:
    messages = full_history[-max_messages:]
    filtered_messages = [m for m in messages if m.content]
    mapped_messages = []

    for m in filtered_messages:
        m.content = re.sub(r"<sup>\d+</sup>", "", m.content or "")
        m.role = (
            ChatMessageRole.ASSISTANT
            if m.role == ChatMessageRole.ASSISTANT
            else ChatMessageRole.USER
        )
        mapped_messages.append(m)

    return pick_messages_in_reverse_for_token_window(
        messages=mapped_messages,
        limit=max_tokens,
        model_info=model_info,
    )


def map_to_chat_messages(messages: list[dict]) -> list[ChatMessage]:
    return [ChatMessage(**msg) for msg in messages]


def pick_messages_in_reverse_for_token_window(
    messages: list[ChatMessage],
    limit: int,
    model_info: LanguageModelInfo | None = None,
) -> list[ChatMessage]:
    if len(messages) < 1 or limit < 1:
        return []

    last_index = len(messages) - 1
    token_count = count_tokens(messages[last_index].content or "", model=model_info)
    while token_count > limit:
        logger.debug(
            f"Limit too low for the initial message. Last message TokenCount {token_count} available tokens {limit} - cutting message in half until it fits",
        )
        content = messages[last_index].content or ""
        messages[last_index].content = content[: len(content) // 2] + "..."
        token_count = count_tokens(messages[last_index].content or "", model=model_info)

    while token_count <= limit and last_index > 0:
        token_count = count_tokens(
            "".join([msg.content or "" for msg in messages[:last_index]]),
            model=model_info,
        )
        if token_count <= limit:
            last_index -= 1

    last_index = max(0, last_index)
    return messages[last_index:]


def list_messages(
    event_user_id: str,
    event_company_id: str,
    chat_id: str,
) -> ListObject[unique_sdk.Message]:
    try:
        messages = unique_sdk.Message.list(
            user_id=event_user_id,
            company_id=event_company_id,
            chatId=chat_id,
        )
        return messages
    except Exception as e:
        logger.error(f"Failed to list chat history: {e}")
        raise e


async def list_messages_async(
    event_user_id: str,
    event_company_id: str,
    chat_id: str,
) -> ListObject[unique_sdk.Message]:
    try:
        messages = await unique_sdk.Message.list_async(
            user_id=event_user_id,
            company_id=event_company_id,
            chatId=chat_id,
        )
        return messages
    except Exception as e:
        logger.error(f"Failed to list chat history: {e}")
        raise e


def get_full_history(
    event_user_id: str,
    event_company_id: str,
    event_payload_chat_id: str,
) -> list[ChatMessage]:
    messages = list_messages(event_user_id, event_company_id, event_payload_chat_id)
    messages = filter_valid_messages(messages)

    return map_to_chat_messages(messages)


async def get_full_history_async(
    event_user_id: str,
    event_company_id: str,
    event_payload_chat_id: str,
) -> list[ChatMessage]:
    messages = await list_messages_async(
        event_user_id,
        event_company_id,
        event_payload_chat_id,
    )
    messages = filter_valid_messages(messages)

    return map_to_chat_messages(messages)


def filter_valid_messages(
    messages: ListObject[unique_sdk.Message],
) -> list[dict[str, Any]]:
    SYSTEM_MESSAGE_PREFIX = "[SYSTEM] "
    roles_to_filter = [ChatMessageRole.SYSTEM.value.lower()]

    # Remove the last two messages
    messages = messages["data"][:-2]  # type: ignore
    filtered_messages = []
    for message in messages:
        if (
            message["text"] is None
            or SYSTEM_MESSAGE_PREFIX in message["text"]
            or message["role"].lower() in roles_to_filter
        ):
            continue
        filtered_messages.append(message)

    return filtered_messages


def create_message_assessment(
    user_id: str,
    company_id: str,
    assistant_message_id: str,
    status: ChatMessageAssessmentStatus,
    type: ChatMessageAssessmentType,
    title: str | None = None,
    explanation: str | None = None,
    label: ChatMessageAssessmentLabel | None = None,
    is_visible: bool = True,
) -> ChatMessageAssessment:
    """Creates a message assessment for an assistant message synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        assistant_message_id (str): The ID of the assistant message to assess
        status (ChatMessageAssessmentStatus): The status of the assessment (e.g. "DONE")
        type (ChatMessageAssessmentType): The type of assessment (e.g. "HALLUCINATION")
        title (str | None): The title of the assessment
        explanation (str | None): Explanation of the assessment
        label (ChatMessageAssessmentLabel | None): The assessment label (e.g. "NEGATIVE")
        is_visible (bool): Whether the assessment is visible to users. Defaults to True.

    Returns:
        ChatMessageAssessment: The created message assessment

    Raises:
        Exception: If the creation fails

    """
    try:
        assessment = unique_sdk.MessageAssessment.create(
            user_id=user_id,
            company_id=company_id,
            messageId=assistant_message_id,
            status=status.name,
            explanation=explanation,
            label=label.name if label else None,
            title=title,
            type=type.name,
            isVisible=is_visible,
        )
        return ChatMessageAssessment(**assessment)
    except Exception as e:
        logger.error(f"Failed to create message assessment: {e}")
        raise e


async def create_message_assessment_async(
    user_id: str,
    company_id: str,
    assistant_message_id: str,
    status: ChatMessageAssessmentStatus,
    type: ChatMessageAssessmentType,
    title: str | None = None,
    explanation: str | None = None,
    label: ChatMessageAssessmentLabel | None = None,
    is_visible: bool = True,
) -> ChatMessageAssessment:
    """Creates a message assessment for an assistant message asynchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        assistant_message_id (str): The ID of the assistant message to assess
        status (ChatMessageAssessmentStatus): The status of the assessment (e.g. "DONE")
        type (ChatMessageAssessmentType): The type of assessment (e.g. "HALLUCINATION")
        title (str | None): The title of the assessment
        explanation (str | None): Explanation of the assessment
        label (ChatMessageAssessmentLabel | None): The assessment label (e.g. "NEGATIVE")
        is_visible (bool): Whether the assessment is visible to users. Defaults to True.

    Returns:
        MessageAssessment: The created message assessment

    Raises:
        Exception: If the creation fails

    """
    try:
        assessment = await unique_sdk.MessageAssessment.create_async(
            user_id=user_id,
            company_id=company_id,
            messageId=assistant_message_id,
            status=status.name,
            explanation=explanation,
            label=label.name if label else None,
            title=title,
            type=type.name,
            isVisible=is_visible,
        )
        return ChatMessageAssessment(**assessment)
    except Exception as e:
        logger.error(f"Failed to create message assessment: {e}")
        raise e


def modify_message_assessment(
    user_id: str,
    company_id: str,
    assistant_message_id: str,
    status: ChatMessageAssessmentStatus,
    type: ChatMessageAssessmentType,
    title: str | None = None,
    explanation: str | None = None,
    label: ChatMessageAssessmentLabel | None = None,
) -> ChatMessageAssessment:
    """Modifies a message assessment for an assistant message synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        assistant_message_id (str): The ID of the assistant message to assess
        status (MessageAssessmentStatus): The status of the assessment (e.g. "DONE")
        title (str | None): The title of the assessment
        explanation (str | None): Explanation of the assessment
        label (ChatMessageAssessmentLabel | None): The assessment label (e.g. "NEGATIVE")
        type (ChatMessageAssessmentType): The type of assessment (e.g. "HALLUCINATION")

    Returns:
        dict: The modified message assessment

    Raises:
        Exception: If the modification fails

    """
    try:
        assessment = unique_sdk.MessageAssessment.modify(
            user_id=user_id,
            company_id=company_id,
            messageId=assistant_message_id,
            status=status.name,
            title=title,
            explanation=explanation,
            label=label.name if label else None,
            type=type.name,
        )
        return ChatMessageAssessment(**assessment)
    except Exception as e:
        logger.error(f"Failed to modify message assessment: {e}")
        raise e


async def modify_message_assessment_async(
    user_id: str,
    company_id: str,
    assistant_message_id: str,
    type: ChatMessageAssessmentType,
    title: str | None = None,
    status: ChatMessageAssessmentStatus | None = None,
    explanation: str | None = None,
    label: ChatMessageAssessmentLabel | None = None,
) -> ChatMessageAssessment:
    """Modifies a message assessment for an assistant message asynchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        assistant_message_id (str): The ID of the assistant message to assess
        type (MessageAssessmentType): The type of assessment (e.g. "HALLUCINATION")
        title (str | None): The title of the assessment
        status (MessageAssessmentStatus): The status of the assessment (e.g. "DONE")
        explanation (str | None): Explanation of the assessment
        label (MessageAssessmentLabel | None): The assessment label (e.g. "NEGATIVE")

    Returns:
        MessageAssessment: The modified message assessment

    Raises:
        Exception: If the modification fails

    """
    try:
        assessment = await unique_sdk.MessageAssessment.modify_async(
            user_id=user_id,
            company_id=company_id,
            messageId=assistant_message_id,
            status=status.name if status else None,
            title=title,
            explanation=explanation,
            label=label.name if label else None,
            type=type.name,
        )
        return ChatMessageAssessment(**assessment)
    except Exception as e:
        logger.error(f"Failed to modify message assessment: {e}")
        raise e


@deprecated("Use stream_complete_with_references instead")
def stream_complete_to_chat(
    company_id: str,
    user_id: str,
    assistant_message_id: str,
    user_message_id: str,
    chat_id: str,
    assistant_id: str,
    messages: LanguageModelMessages | list[ChatCompletionMessageParam],
    model_name: LanguageModelName | str,
    content_chunks: list[ContentChunk] | None = None,
    debug_info: dict = {},
    temperature: float = DEFAULT_COMPLETE_TEMPERATURE,
    timeout: int = DEFAULT_COMPLETE_TIMEOUT,
    tools: list[LanguageModelTool | LanguageModelToolDescription] | None = None,
    start_text: str | None = None,
    other_options: dict | None = None,
) -> LanguageModelStreamResponse:
    return stream_complete_with_references(
        company_id=company_id,
        user_id=user_id,
        assistant_message_id=assistant_message_id,
        user_message_id=user_message_id,
        chat_id=chat_id,
        assistant_id=assistant_id,
        messages=messages,
        model_name=model_name,
        content_chunks=content_chunks,
        debug_info=debug_info,
        temperature=temperature,
        timeout=timeout,
        tools=tools,
        start_text=start_text,
        other_options=other_options,
    )


def stream_complete_with_references(
    company_id: str,
    user_id: str,
    assistant_message_id: str,
    user_message_id: str,
    chat_id: str,
    assistant_id: str,
    messages: LanguageModelMessages | list[ChatCompletionMessageParam],
    model_name: LanguageModelName | str,
    content_chunks: list[ContentChunk] | None = None,
    debug_info: dict | None = None,
    temperature: float = DEFAULT_COMPLETE_TEMPERATURE,
    timeout: int = DEFAULT_COMPLETE_TIMEOUT,
    tools: Sequence[LanguageModelTool | LanguageModelToolDescription] | None = None,
    start_text: str | None = None,
    tool_choice: ChatCompletionToolChoiceOptionParam | None = None,
    other_options: dict | None = None,
) -> LanguageModelStreamResponse:
    """Streams a completion synchronously.

    Args:
        company_id (str): The company ID associated with the request.
        user_id (str): The user ID for the request.
        assistant_message_id (str): The assistant message ID.
        user_message_id (str): The user message ID.
        chat_id (str): The chat ID.
        assistant_id (str): The assistant ID.
        messages (LanguageModelMessages): The messages to complete.
        model_name (LanguageModelName | str): The model name.
        content_chunks (list[ContentChunk]): Content chunks for context.
        debug_info (dict): Debug information.
        temperature (float): Temperature setting.
        timeout (int): Timeout in milliseconds.
        tools (Optional[list[LanguageModelTool | LanguageModelToolDescription ]]): Optional tools.
        start_text (Optional[str]): Starting text.
        other_options (Optional[dict]): Additional options.

    Returns:
        LanguageModelStreamResponse: The streaming response object.

    """
    options, model, messages_dict, search_context = (
        _prepare_all_completions_params_util(
            messages=messages,
            model_name=model_name,
            temperature=temperature,
            tools=tools,
            other_options=other_options,
            tool_choice=tool_choice,
            content_chunks=content_chunks or [],
        )
    )

    try:
        response = unique_sdk.Integrated.chat_stream_completion(
            user_id=user_id,
            company_id=company_id,
            assistantMessageId=assistant_message_id,
            userMessageId=user_message_id,
            messages=messages_dict,
            chatId=chat_id,
            searchContext=search_context,
            model=model,
            timeout=timeout,
            assistantId=assistant_id,
            debugInfo=debug_info or {},
            options=options,  # type: ignore
            startText=start_text,
        )
        return LanguageModelStreamResponse(**response)
    except Exception as e:
        logger.error(f"Error streaming completion: {e}")
        raise e


@deprecated("Use stream_complete_with_references_async instead")
async def stream_complete_to_chat_async(
    company_id: str,
    user_id: str,
    assistant_message_id: str,
    user_message_id: str,
    chat_id: str,
    assistant_id: str,
    messages: LanguageModelMessages | list[ChatCompletionMessageParam],
    model_name: LanguageModelName | str,
    content_chunks: list[ContentChunk] | None = None,
    debug_info: dict = {},
    temperature: float = DEFAULT_COMPLETE_TEMPERATURE,
    timeout: int = DEFAULT_COMPLETE_TIMEOUT,
    tools: list[LanguageModelTool | LanguageModelToolDescription] | None = None,
    start_text: str | None = None,
    other_options: dict | None = None,
) -> LanguageModelStreamResponse:
    return await stream_complete_with_references_async(
        company_id=company_id,
        user_id=user_id,
        assistant_message_id=assistant_message_id,
        user_message_id=user_message_id,
        chat_id=chat_id,
        assistant_id=assistant_id,
        messages=messages,
        model_name=model_name,
        content_chunks=content_chunks,
        debug_info=debug_info,
        temperature=temperature,
        timeout=timeout,
        tools=tools,
        start_text=start_text,
        other_options=other_options,
    )


async def stream_complete_with_references_async(
    company_id: str,
    user_id: str,
    assistant_message_id: str,
    user_message_id: str,
    chat_id: str,
    assistant_id: str,
    messages: LanguageModelMessages | list[ChatCompletionMessageParam],
    model_name: LanguageModelName | str,
    content_chunks: list[ContentChunk] | None = None,
    debug_info: dict | None = None,
    temperature: float = DEFAULT_COMPLETE_TEMPERATURE,
    timeout: int = DEFAULT_COMPLETE_TIMEOUT,
    tools: Sequence[LanguageModelTool | LanguageModelToolDescription] | None = None,
    tool_choice: ChatCompletionToolChoiceOptionParam | None = None,
    start_text: str | None = None,
    other_options: dict | None = None,
) -> LanguageModelStreamResponse:
    """Streams a completion asynchronously.

    Args: [same as stream_complete]

    Returns:
        LanguageModelStreamResponse: The streaming response object.

    """
    options, model, messages_dict, search_context = (
        _prepare_all_completions_params_util(
            messages=messages,
            model_name=model_name,
            temperature=temperature,
            tools=tools,
            tool_choice=tool_choice,
            other_options=other_options,
            content_chunks=content_chunks or [],
        )
    )

    try:
        response = await unique_sdk.Integrated.chat_stream_completion_async(
            user_id=user_id,
            company_id=company_id,
            assistantMessageId=assistant_message_id,
            userMessageId=user_message_id,
            messages=messages_dict,
            chatId=chat_id,
            searchContext=search_context,
            model=model,
            timeout=timeout,
            assistantId=assistant_id,
            debugInfo=debug_info or {},
            options=options,  # type: ignore
            startText=start_text,
        )
        return LanguageModelStreamResponse(**response)
    except Exception as e:
        logger.error(f"Error streaming completion: {e}")
        raise e


def _get_model_dump_or_none(model: BaseModel | None) -> dict | None:
    if model is None:
        return None
    return model.model_dump()


def create_message_log(
    user_id: str,
    company_id: str,
    message_id: str,
    text: str,
    status: MessageLogStatus,
    order: int,
    details: MessageLogDetails | None = None,
    uncited_references: MessageLogUncitedReferences | None = None,
    references: list[ContentReference] | None = None,
) -> MessageLog:
    """Creates a message log synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_id (str): The ID of the message this log belongs to.
        text (str): The log text content.
        status (MessageLogStatus): The status of this log entry.
        order (int): The order/sequence number of this log entry.
        details (MessageLogDetails | None): Additional details about this log entry.
        uncited_references (MessageLogUncitedReferences | None): References that are not cited.
        references (list[ContentReference] | None): List of references for this log.

    Returns:
        MessageLog: The created message log.

    Raises:
        Exception: If the creation fails.

    """
    try:
        references_list = (
            map_references_with_original_index(references) if references else None
        )
        message_log = unique_sdk.MessageLog.create(
            user_id=user_id,
            company_id=company_id,
            messageId=message_id,
            text=text,
            status=status.value,
            order=order,
            details=_get_model_dump_or_none(details),
            uncitedReferences=_get_model_dump_or_none(uncited_references),
            references=references_list,  # type: ignore
        )
        return MessageLog(**message_log)
    except Exception as e:
        logger.error(f"Failed to create message log: {e}")
        raise e


async def create_message_log_async(
    user_id: str,
    company_id: str,
    message_id: str,
    text: str,
    status: MessageLogStatus,
    order: int,
    details: MessageLogDetails | None = None,
    uncited_references: MessageLogUncitedReferences | None = None,
    references: list[ContentReference] | None = None,
) -> MessageLog:
    """Creates a message log asynchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_id (str): The ID of the message this log belongs to.
        text (str): The log text content.
        status (MessageLogStatus): The status of this log entry.
        order (int): The order/sequence number of this log entry.
        details (MessageLogDetails | None): Additional details about this log entry.
        uncited_references (MessageLogUncitedReferences | None): References that are not cited.
        references (list[ContentReference] | None): List of references for this log.

    Returns:
        MessageLog: The created message log.

    Raises:
        Exception: If the creation fails.

    """
    try:
        references_list = (
            map_references_with_original_index(references) if references else None
        )
        message_log = await unique_sdk.MessageLog.create_async(
            user_id=user_id,
            company_id=company_id,
            messageId=message_id,
            text=text,
            status=status.value,
            order=order,
            details=_get_model_dump_or_none(details),
            uncitedReferences=_get_model_dump_or_none(uncited_references),
            references=references_list,  # type: ignore
        )
        return MessageLog(**message_log)
    except Exception as e:
        logger.error(f"Failed to create message log: {e}")
        raise e


def update_message_log(
    user_id: str,
    company_id: str,
    message_log_id: str,
    order: int,
    text: str | None = None,
    status: MessageLogStatus | None = None,
    details: MessageLogDetails | None = None,
    uncited_references: MessageLogUncitedReferences | None = None,
    references: list[ContentReference] | None = None,
) -> MessageLog:
    """Updates a message log synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_log_id (str): The ID of the message log to update.
        order (int): The order/sequence number (required).
        text (str | None): The updated log text content.
        status (MessageLogStatus | None): The updated status.
        details (MessageLogDetails | None): Updated additional details.
        uncited_references (MessageLogUncitedReferences | None): Updated uncited references.
        references (list[ContentReference] | None): Updated list of references.

    Returns:
        MessageLog: The updated message log.

    Raises:
        Exception: If the update fails.

    """
    try:
        references_list = (
            map_references_with_original_index(references) if references else None
        )
        message_log = unique_sdk.MessageLog.update(
            user_id=user_id,
            company_id=company_id,
            message_log_id=message_log_id,
            text=text,
            status=status.value if status else None,
            order=order,
            details=_get_model_dump_or_none(details),
            uncitedReferences=_get_model_dump_or_none(uncited_references),
            references=references_list,  # type: ignore
        )
        return MessageLog(**message_log)
    except Exception as e:
        logger.error(f"Failed to update message log: {e}")
        raise e


async def update_message_log_async(
    user_id: str,
    company_id: str,
    message_log_id: str,
    order: int,
    text: str | None = None,
    status: MessageLogStatus | None = None,
    details: MessageLogDetails | None = None,
    uncited_references: MessageLogUncitedReferences | None = None,
    references: list[ContentReference] | None = None,
) -> MessageLog:
    """Updates a message log asynchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_log_id (str): The ID of the message log to update.
        order (int): The order/sequence number (required).
        text (str | None): The updated log text content.
        status (MessageLogStatus | None): The updated status.
        details (MessageLogDetails | None): Updated additional details.
        uncited_references (MessageLogUncitedReferences | None): Updated uncited references.
        references (list[ContentReference] | None): Updated list of references.

    Returns:
        MessageLog: The updated message log.

    Raises:
        Exception: If the update fails.

    """
    try:
        references_list = (
            map_references_with_original_index(references) if references else None
        )
        message_log = await unique_sdk.MessageLog.update_async(
            user_id=user_id,
            company_id=company_id,
            message_log_id=message_log_id,
            text=text,
            status=status.value if status else None,
            order=order,
            details=_get_model_dump_or_none(details),
            uncitedReferences=_get_model_dump_or_none(uncited_references),
            references=references_list,  # type: ignore
        )
        return MessageLog(**message_log)
    except Exception as e:
        logger.error(f"Failed to update message log: {e}")
        raise e


def create_message_execution(
    user_id: str,
    company_id: str,
    message_id: str,
    chat_id: str,
    type: MessageExecutionType = MessageExecutionType.DEEP_RESEARCH,
    seconds_remaining: int | None = None,
    percentage_completed: int | None = None,
    is_queueable: bool = True,
    execution_options: dict | None = None,
    progress_title: str | None = None,
) -> MessageExecution:
    """Creates a message execution synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_id (str): The ID of the message this execution belongs to.
        chat_id (str): The chat ID.
        type (MessageExecutionType): The type of execution. Defaults to DEEP_RESEARCH.
        seconds_remaining (int | None): Estimated seconds remaining for completion.
        percentage_completed (int | None): Percentage of completion (0-100).
        is_queueable (bool): Whether the execution is queueable. Defaults to True. If true, then the progress will be updated in the background by the execution pipeline. Set to False if you want to update the progress manually.
        execution_options (dict | None): Additional execution options. Defaults to None.
        progress_title (str | None): The title of the progress bar. If not provided, the title of the last message log is taken.

    Returns:
        MessageExecution: The created message execution.

    Raises:
        Exception: If the creation fails.

    """
    try:
        message_execution = unique_sdk.MessageExecution.create(
            user_id=user_id,
            company_id=company_id,
            messageId=message_id,
            chatId=chat_id,
            type=type.value,
            secondsRemaining=seconds_remaining,
            percentageCompleted=percentage_completed,
            isQueueable=is_queueable,
            executionOptions=execution_options,
            progressTitle=progress_title,
        )
        return MessageExecution(**message_execution)
    except Exception as e:
        logger.error(f"Failed to create message execution: {e}")
        raise e


async def create_message_execution_async(
    user_id: str,
    company_id: str,
    message_id: str,
    chat_id: str,
    type: MessageExecutionType = MessageExecutionType.DEEP_RESEARCH,
    seconds_remaining: int | None = None,
    percentage_completed: int | None = None,
    is_queueable: bool = True,
    execution_options: dict | None = None,
    progress_title: str | None = None,
) -> MessageExecution:
    """Creates a message execution asynchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_id (str): The ID of the message this execution belongs to.
        chat_id (str): The chat ID.
        type (MessageExecutionType): The type of execution. Defaults to DEEP_RESEARCH.
        seconds_remaining (int | None): Estimated seconds remaining for completion.
        percentage_completed (int | None): Percentage of completion (0-100).
        is_queueable (bool): Whether the execution is queueable. Defaults to True. If true, then the progress will be updated in the background by the execution pipeline. Set to False if you want to update the progress manually.
        execution_options (dict | None): Additional execution options. Defaults to None.
        progress_title (str | None): The title of the progress bar. If not provided, the title of the last message log is taken.

    Returns:
        MessageExecution: The created message execution.

    Raises:
        Exception: If the creation fails.

    """
    try:
        message_execution = await unique_sdk.MessageExecution.create_async(
            user_id=user_id,
            company_id=company_id,
            messageId=message_id,
            chatId=chat_id,
            type=type.value,
            secondsRemaining=seconds_remaining,
            percentageCompleted=percentage_completed,
            isQueueable=is_queueable,
            executionOptions=execution_options,
            progressTitle=progress_title,
        )
        return MessageExecution(**message_execution)
    except Exception as e:
        logger.error(f"Failed to create message execution: {e}")
        raise e


def get_message_execution(
    user_id: str,
    company_id: str,
    message_id: str,
) -> MessageExecution:
    """Gets a message execution by message ID synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_id (str): The ID of the message to get execution for.

    Returns:
        MessageExecution: The message execution.

    Raises:
        Exception: If the retrieval fails.

    """
    try:
        message_execution = unique_sdk.MessageExecution.get(
            user_id=user_id,
            company_id=company_id,
            messageId=message_id,
        )
        return MessageExecution(**message_execution)
    except Exception as e:
        logger.error(f"Failed to get message execution: {e}")
        raise e


async def get_message_execution_async(
    user_id: str,
    company_id: str,
    message_id: str,
) -> MessageExecution:
    """Gets a message execution by message ID asynchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_id (str): The ID of the message to get execution for.

    Returns:
        MessageExecution: The message execution.

    Raises:
        Exception: If the retrieval fails.

    """
    try:
        message_execution = await unique_sdk.MessageExecution.get_async(
            user_id=user_id,
            company_id=company_id,
            messageId=message_id,
        )
        return MessageExecution(**message_execution)
    except Exception as e:
        logger.error(f"Failed to get message execution: {e}")
        raise e


def update_message_execution(
    user_id: str,
    company_id: str,
    message_id: str,
    status: MessageExecutionUpdateStatus | None = None,
    seconds_remaining: int | None = None,
    percentage_completed: int | None = None,
    progress_title: str | None = None,
) -> MessageExecution:
    """Updates a message execution synchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_id (str): The ID of the message to update execution for.
        status (MessageExecutionUpdateStatus | None): The updated status (COMPLETED or FAILED). Defaults to None.
        seconds_remaining (int | None): Updated estimated seconds remaining.
        percentage_completed (int | None): Updated percentage of completion (0-100).
        progress_title (str | None): The title of the progress bar. If not provided, the title of the last message log is taken.

    Returns:
        MessageExecution: The updated message execution.

    Raises:
        Exception: If the update fails.

    """
    try:
        status_value = status.value if status else None
        message_execution = unique_sdk.MessageExecution.update(
            user_id=user_id,
            company_id=company_id,
            messageId=message_id,
            status=status_value,
            secondsRemaining=seconds_remaining,
            percentageCompleted=percentage_completed,
            progressTitle=progress_title,
        )
        return MessageExecution(**message_execution)
    except Exception as e:
        logger.error(f"Failed to update message execution: {e}")
        raise e


async def update_message_execution_async(
    user_id: str,
    company_id: str,
    message_id: str,
    status: MessageExecutionUpdateStatus | None = None,
    seconds_remaining: int | None = None,
    percentage_completed: int | None = None,
    progress_title: str | None = None,
) -> MessageExecution:
    """Updates a message execution asynchronously.

    Args:
        user_id (str): The user ID.
        company_id (str): The company ID.
        message_id (str): The ID of the message to update execution for.
        status (MessageExecutionUpdateStatus | None): The updated status (COMPLETED or FAILED). Defaults to None.
        seconds_remaining (int | None): Updated estimated seconds remaining.
        percentage_completed (int | None): Updated percentage of completion (0-100).
        progress_title (str | None): The title of the progress bar. If not provided, the title of the last message log is taken.

    Returns:
        MessageExecution: The updated message execution.

    Raises:
        Exception: If the update fails.

    """
    try:
        status_value = status.value if status else None
        message_execution = await unique_sdk.MessageExecution.update_async(
            user_id=user_id,
            company_id=company_id,
            messageId=message_id,
            status=status_value,
            secondsRemaining=seconds_remaining,
            percentageCompleted=percentage_completed,
            progressTitle=progress_title,
        )
        return MessageExecution(**message_execution)
    except Exception as e:
        logger.error(f"Failed to update message execution: {e}")
        raise e



from typing import TYPE_CHECKING, Callable

from unique_toolkit.agentic.tools.schemas import BaseToolConfig
from unique_toolkit.agentic.tools.tool import Tool

if TYPE_CHECKING:
    from unique_toolkit.agentic.tools.config import ToolBuildConfig


class ToolFactory:
    tool_map: dict[str, type[Tool]] = {}
    tool_config_map: dict[str, Callable] = {}

    @classmethod
    def register_tool_config(cls, tool_name: str, tool_config: type[BaseToolConfig]):
        cls.tool_config_map[tool_name] = tool_config

    @classmethod
    def register_tool(
        cls,
        tool: type[Tool],
        tool_config: type[BaseToolConfig],
    ):
        cls.tool_map[tool.name] = tool
        cls.tool_config_map[tool.name] = tool_config

    @classmethod
    def build_tool(cls, tool_name: str, *args, **kwargs) -> Tool[BaseToolConfig]:
        tool = cls.tool_map[tool_name](*args, **kwargs)
        return tool

    @classmethod
    def build_tool_with_settings(
        cls, tool_name: str, settings: "ToolBuildConfig", *args, **kwargs
    ) -> Tool[BaseToolConfig]:
        tool = cls.tool_map[tool_name](*args, **kwargs)
        tool.settings = settings
        return tool

    @classmethod
    def build_tool_config(cls, tool_name: str, **kwargs) -> BaseToolConfig:
        if tool_name not in cls.tool_config_map:
            raise ValueError(f"Tool {tool_name} not found")
        return cls.tool_config_map[tool_name](**kwargs)



"""OpenAI framework utilities."""

from .client import get_async_openai_client, get_openai_client
from .message_builder import OpenAIMessageBuilder

__all__ = ["get_openai_client", "OpenAIMessageBuilder", "get_async_openai_client"]



from datetime import datetime
from enum import StrEnum
from typing import Any

from humps import camelize
from pydantic import AnyUrl, BaseModel, ConfigDict, Field, field_serializer

# set config to convert camelCase to snake_case
model_config = ConfigDict(
    alias_generator=camelize,
    populate_by_name=True,
)


class ElicitationObject(StrEnum):
    ELICITATION = "elicitation"


class ElicitationMode(StrEnum):
    FORM = "FORM"
    URL = "URL"


class ElicitationAction(StrEnum):
    ACCEPT = "ACCEPT"
    DECLINE = "DECLINE"
    CANCEL = "CANCEL"


class ElicitationStatus(StrEnum):
    PENDING = "PENDING"
    ACCEPTED = "ACCEPTED"
    DECLINED = "DECLINED"
    CANCELLED = "CANCELLED"
    EXPIRED = "EXPIRED"


class ElicitationSource(StrEnum):
    API = "INTERNAL_TOOL"
    MCP = "MCP_SERVER"


class Elicitation(BaseModel):
    """
    Represents an elicitation request.
    """

    model_config = model_config

    id: str = Field(description="Unique identifier for the elicitation")
    object: ElicitationObject = Field(description="Object type, always 'elicitation'")
    source: ElicitationSource = Field(
        description="Source of the elicitation (API or MCP)"
    )
    mode: ElicitationMode = Field(description="Elicitation mode: FORM or URL")
    status: ElicitationStatus = Field(
        description="Current status of the elicitation request"
    )
    message: str = Field(description="Message to display to the user")
    mcp_server_id: str | None = Field(
        default=None, description="MCP server ID if elicitation is from MCP"
    )
    tool_name: str | None = Field(
        default=None, description="Name of the tool requesting elicitation"
    )
    json_schema: dict[str, Any] | None = Field(
        default=None,
        validation_alias="schema",
        serialization_alias="schema",
        description="JSON schema for FORM mode elicitation",
    )
    url: AnyUrl | None = Field(default=None, description="URL for URL mode elicitation")
    external_elicitation_id: str | None = Field(
        default=None, description="External elicitation ID for tracking"
    )
    response_content: dict[str, Any] | None = Field(
        default=None, description="Content of the user's response"
    )
    responded_at: datetime | None = Field(
        default=None, description="Timestamp when the elicitation was responded to"
    )
    company_id: str = Field(description="Company ID")
    user_id: str = Field(description="User ID")
    chat_id: str | None = Field(
        default=None, description="Chat ID if associated with a chat"
    )
    message_id: str | None = Field(
        default=None, description="Message ID if associated with a message"
    )
    metadata: dict[str, Any] | None = Field(
        default=None, description="Additional metadata"
    )
    created_at: datetime = Field(
        description="Timestamp when the elicitation was created"
    )
    updated_at: datetime | None = Field(
        default=None, description="Timestamp when the elicitation was last updated"
    )
    expires_at: datetime | None = Field(
        default=None, description="Timestamp when the elicitation expires"
    )

    @field_serializer("created_at", "updated_at", "expires_at", "responded_at")
    def serialize_datetime(self, value: datetime | None) -> str | None:
        """Serialize datetime to ISO format string."""
        if value is None:
            return None
        return value.isoformat()


class ElicitationResponseResult(BaseModel):
    """
    Response for responding to an elicitation request.
    """

    model_config = model_config

    success: bool = Field(description="Whether the response was successful")
    message: str | None = Field(
        default=None, description="Optional message about the response result"
    )


class ElicitationList(BaseModel):
    """
    Response for getting pending elicitations.
    """

    model_config = model_config

    elicitations: list[Elicitation] = Field(description="List of elicitation requests")


class CreateElicitationParams(BaseModel):
    """
    Parameters for creating an elicitation request.
    """

    model_config = model_config

    mode: ElicitationMode = Field(description="Elicitation mode")
    message: str = Field(description="Message to display to the user")
    tool_name: str = Field(description="Name of the tool requesting elicitation")
    json_schema: dict[str, Any] | None = Field(
        default=None,
        validation_alias="schema",
        serialization_alias="schema",
        description="JSON schema for FORM mode elicitation",
    )
    url: AnyUrl | None = Field(default=None, description="URL for URL mode elicitation")
    external_elicitation_id: str | None = Field(
        default=None, description="External elicitation ID for tracking"
    )
    chat_id: str | None = Field(
        default=None, description="Chat ID if associated with a chat"
    )
    message_id: str | None = Field(
        default=None, description="Message ID if associated with a message"
    )
    expires_in_seconds: int | None = Field(
        default=None, description="Expiration time in seconds"
    )
    metadata: dict[str, Any] | None = Field(
        default=None, description="Additional metadata"
    )


class RespondToElicitationParams(BaseModel):
    """
    Parameters for responding to an elicitation request.
    """

    model_config = model_config

    elicitation_id: str = Field(description="The elicitation ID to respond to")
    action: ElicitationAction = Field(description="Action to take on the elicitation")
    content: dict[str, str | int | bool | list[str]] | None = Field(
        default=None, description="Response content for ACCEPT action"
    )



from .constants import DOMAIN_NAME as DOMAIN_NAME
from .schemas import (
    Content as Content,
)
from .schemas import (
    ContentChunk as ContentChunk,
)
from .schemas import (
    ContentMetadata as ContentMetadata,
)
from .schemas import (
    ContentReference as ContentReference,
)
from .schemas import (
    ContentRerankerConfig as ContentRerankerConfig,
)
from .schemas import (
    ContentSearchResult as ContentSearchResult,
)
from .schemas import (
    ContentSearchType as ContentSearchType,
)
from .schemas import (
    ContentUploadInput as ContentUploadInput,
)
from .service import ContentService as ContentService
from .utils import (
    count_tokens as count_tokens,
)
from .utils import (
    merge_content_chunks as merge_content_chunks,
)
from .utils import (
    pick_content_chunks_for_token_window as pick_content_chunks_for_token_window,
)
from .utils import (
    sort_content_chunks as sort_content_chunks,
)



import asyncio
import logging
import re
from mimetypes import guess_type
from typing import NamedTuple, override

from openai import AsyncOpenAI
from openai.types.responses.response_output_text import AnnotationContainerFileCitation
from pydantic import BaseModel, Field, RootModel
from pydantic.json_schema import SkipJsonSchema

from unique_toolkit import ChatService
from unique_toolkit._common.execution import failsafe_async
from unique_toolkit.agentic.feature_flags.feature_flags import feature_flags
from unique_toolkit.agentic.postprocessor.postprocessor_manager import (
    ResponsesApiPostprocessor,
)
from unique_toolkit.agentic.short_term_memory_manager.persistent_short_term_memory_manager import (
    PersistentShortMemoryManager,
)
from unique_toolkit.agentic.tools.config import get_configuration_dict
from unique_toolkit.content.schemas import ContentReference
from unique_toolkit.content.service import ContentService
from unique_toolkit.language_model.schemas import ResponsesLanguageModelStreamResponse
from unique_toolkit.services.knowledge_base import KnowledgeBaseService
from unique_toolkit.short_term_memory.service import ShortTermMemoryService

logger = logging.getLogger(__name__)


class DisplayCodeInterpreterFilesPostProcessorConfig(BaseModel):
    model_config = get_configuration_dict()
    upload_to_chat: SkipJsonSchema[bool] = Field(
        default=True,
        description="Whether to upload the generated files to the chat.",
    )
    upload_scope_id: SkipJsonSchema[str] = Field(
        default="<SCOPE_ID_PLACEHOLDER>",
        description="The scope ID where the generated files will be uploaded. Ignored if `uploadToChat` is set.",
    )

    file_download_failed_message: str = Field(
        default="⚠️ File download failed ...",
        description="The message to display when a file download fails.",
    )
    max_concurrent_file_downloads: int = Field(
        default=10,
        description="The maximum number of concurrent file downloads.",
    )


class _ContentInfo(NamedTuple):
    filename: str
    content_id: str


_SHORT_TERM_MEMORY_KEY = "code_interpreter_files"
_DisplayedFilesShortTermMemorySchema = RootModel[list[_ContentInfo]]


def _init_short_term_memory_manager(
    company_id: str, user_id: str, chat_id: str
) -> PersistentShortMemoryManager[_DisplayedFilesShortTermMemorySchema]:
    short_term_memory_service = ShortTermMemoryService(
        company_id=company_id,
        user_id=user_id,
        chat_id=chat_id,
        message_id=None,
    )
    return PersistentShortMemoryManager(
        short_term_memory_service=short_term_memory_service,
        short_term_memory_schema=_DisplayedFilesShortTermMemorySchema,
        short_term_memory_name=_SHORT_TERM_MEMORY_KEY,
    )


class DisplayCodeInterpreterFilesPostProcessor(
    ResponsesApiPostprocessor,
):
    def __init__(
        self,
        client: AsyncOpenAI,
        content_service: ContentService | KnowledgeBaseService,
        config: DisplayCodeInterpreterFilesPostProcessorConfig,
        chat_service: ChatService | None = None,
        # Short term memory arguments, we prefer to explicitely pass the required auth variables
        # as it is crucial that we use chat-level short term memory not to leak user files to other chats.
        # Technically, short term memory can be scoped company-level, we would like to ensure this case is avoided.
        company_id: str | None = None,
        user_id: str | None = None,
        chat_id: str | None = None,
    ) -> None:
        super().__init__(self.__class__.__name__)

        self._content_service = content_service
        self._chat_service = chat_service
        self._client = client
        self._config = config
        self._company_id = company_id

        if self._chat_service is None:
            raise ValueError("ChatService is required if uploadToChat is True")

        self._short_term_memory_manager = None
        if chat_id is not None and user_id is not None and company_id is not None:
            self._short_term_memory_manager = _init_short_term_memory_manager(
                company_id=company_id,
                user_id=user_id,
                chat_id=chat_id,
            )

    @override
    async def run(self, loop_response: ResponsesLanguageModelStreamResponse) -> None:
        logger.info("Fetching and adding code interpreter files to the response")

        container_files = loop_response.container_files
        logger.info("Found %s container files", len(container_files))

        self._content_map: dict[str, str | None] = await self._load_previous_files()  # type: ignore

        semaphore = asyncio.Semaphore(self._config.max_concurrent_file_downloads)
        tasks = [
            self._download_and_upload_container_files_to_knowledge_base(
                citation, semaphore
            )
            for citation in container_files
        ]
        results = await asyncio.gather(*tasks)

        for citation, result in zip(container_files, results):
            # Overwrite if file has been re-generated
            self._content_map[citation.filename] = (
                result.content_id if result is not None else None
            )

        await self._save_generated_files(
            [
                _ContentInfo(filename=filename, content_id=content_id)
                for filename, content_id in self._content_map.items()
                if content_id is not None
            ]
        )

    @override
    def apply_postprocessing_to_response(
        self, loop_response: ResponsesLanguageModelStreamResponse
    ) -> bool:
        ref_number = _get_next_ref_number(loop_response.message.references)
        changed = False

        for filename, content_id in self._content_map.items():
            if content_id is None:
                loop_response.message.text, replaced = _replace_container_file_error(
                    text=loop_response.message.text,
                    filename=filename,
                    error_message=self._config.file_download_failed_message,
                )
                changed |= replaced
                continue

            is_image = (guess_type(filename)[0] or "").startswith("image/")
            is_html = (guess_type(filename)[0] or "") == "text/html"

            # Images
            if is_image:
                loop_response.message.text, replaced = (
                    _replace_container_image_citation(
                        text=loop_response.message.text,
                        filename=filename,
                        content_id=content_id,
                    )
                )
                changed |= replaced

            # HTML (behind feature flag)
            elif is_html and feature_flags.enable_html_rendering_un_15131.is_enabled(
                self._company_id
            ):
                loop_response.message.text, replaced = _replace_container_html_citation(
                    text=loop_response.message.text,
                    filename=filename,
                    content_id=content_id,
                )
                changed |= replaced

            # Files (including HTML when feature flag is disabled)
            else:
                loop_response.message.text, replaced = _replace_container_file_citation(
                    text=loop_response.message.text,
                    filename=filename,
                    ref_number=ref_number,
                )
                changed |= replaced

            is_html_rendered = (
                is_html
                and feature_flags.enable_html_rendering_un_15131.is_enabled(
                    self._company_id
                )
            )
            if replaced and not is_image and not is_html_rendered:
                loop_response.message.references.append(
                    ContentReference(
                        sequence_number=ref_number,
                        source_id=content_id,
                        source="node-ingestion-chunks",
                        url=f"unique://content/{content_id}",
                        name=filename,
                    )
                )
                ref_number += 1
        return changed

    @override
    async def remove_from_text(self, text) -> str:
        return text

    @failsafe_async(failure_return_value=None, logger=logger)
    async def _download_and_upload_container_files_to_knowledge_base(
        self,
        container_file: AnnotationContainerFileCitation,
        semaphore: asyncio.Semaphore,
    ) -> _ContentInfo | None:
        async with semaphore:
            logger.info("Fetching file content for %s", container_file.filename)
            file_content = await self._client.containers.files.content.retrieve(
                container_id=container_file.container_id, file_id=container_file.file_id
            )

            logger.info(
                "Uploading file content for %s to knowledge base",
                container_file.filename,
            )

            assert self._chat_service is not None  # Checked in __init__
            content = await self._chat_service.upload_to_chat_from_bytes_async(
                content=file_content.content,
                content_name=container_file.filename,
                mime_type=guess_type(container_file.filename)[0] or "text/plain",
                skip_ingestion=True,
                hide_in_chat=True,
            )

            return _ContentInfo(filename=container_file.filename, content_id=content.id)

    async def _load_previous_files(self) -> dict[str, str]:
        if self._short_term_memory_manager is None:
            return {}

        logger.info(
            "Loading previously generated code interpreter files from short term memory"
        )
        memory = await self._short_term_memory_manager.load_async()

        if memory is None:
            logger.info(
                "No previously generated code interpreter files found in short term memory"
            )
            return {}

        logger.info(
            "Found %s previously generated code interpreter files", len(memory.root)
        )

        return {content.filename: content.content_id for content in memory.root}

    async def _save_generated_files(self, content_infos: list[_ContentInfo]) -> None:
        if self._short_term_memory_manager is None or len(content_infos) == 0:
            return

        await self._short_term_memory_manager.save_async(
            _DisplayedFilesShortTermMemorySchema(root=content_infos)
        )


def _get_next_ref_number(references: list[ContentReference]) -> int:
    max_ref_number = 0
    for ref in references:
        max_ref_number = max(max_ref_number, ref.sequence_number)
    return max_ref_number + 1


def _replace_container_file_error(
    text: str, filename: str, error_message: str
) -> tuple[str, bool]:
    image_markdown = rf"!?\[.*?\]\(sandbox:/mnt/data/{re.escape(filename)}\)"

    if not re.search(image_markdown, text):
        logger.info("No image markdown found for %s", filename)
        return text, False

    logger.info("Displaying image %s", filename)
    return re.sub(
        image_markdown,
        error_message,
        text,
    ), True


def _replace_container_image_citation(
    text: str, filename: str, content_id: str
) -> tuple[str, bool]:
    image_markdown = rf"!?\[.*?\]\(sandbox:/mnt/data/{re.escape(filename)}\)"

    if not re.search(image_markdown, text):
        logger.info("No image markdown found for %s", filename)
        return text, False

    logger.info("Displaying image %s", filename)
    return re.sub(
        image_markdown,
        f"![image](unique://content/{content_id})",
        text,
    ), True


def _replace_container_html_citation(
    text: str, filename: str, content_id: str
) -> tuple[str, bool]:
    html_markdown = rf"!?\[.*?\]\(sandbox:/mnt/data/{re.escape(filename)}\)"

    if not re.search(html_markdown, text):
        logger.info("No HTML markdown found for %s", filename)
        return text, False

    logger.info("Displaying HTML %s", filename)
    html_rendering_block = f"""```HtmlRendering
100%
500px

unique://content/{content_id}

```"""
    return re.sub(
        html_markdown,
        html_rendering_block,
        text,
    ), True


def _replace_container_file_citation(
    text: str, filename: str, ref_number: int
) -> tuple[str, bool]:
    file_markdown = rf"\[.*?\]\(sandbox:/mnt/data/{re.escape(filename)}\)"

    if not re.search(file_markdown, text):
        logger.info("No file markdown found for %s", filename)
        return text, False

    logger.info("Displaying file %s", filename)
    return re.sub(
        file_markdown,
        f"<sup>{ref_number}</sup>",
        text,
    ), True



from pydantic import Field

from unique_toolkit.agentic.tools.a2a.evaluation import SubAgentEvaluationConfig
from unique_toolkit.agentic.tools.a2a.postprocessing import SubAgentDisplayConfig
from unique_toolkit.agentic.tools.a2a.tool import SubAgentToolConfig


# SubAgentToolConfig with display and evaluation configs
class ExtendedSubAgentToolConfig(SubAgentToolConfig):
    response_display_config: SubAgentDisplayConfig = Field(
        default_factory=SubAgentDisplayConfig,
        description="Configuration for how to display the sub-agent response.",
    )
    evaluation_config: SubAgentEvaluationConfig = Field(
        default_factory=SubAgentEvaluationConfig,
        description="Configuration for handling assessments of the sub-agent response.",
    )



import json
import logging
from typing import Any, NamedTuple, Sequence

import unique_sdk
from openai.types.responses import (
    ResponseIncludable,
    ResponseInputItemParam,
    ResponseInputParam,
    ResponseOutputItem,
    ResponseTextConfigParam,
    ToolParam,
    response_create_params,
)
from openai.types.shared_params import Metadata, Reasoning
from pydantic import BaseModel, TypeAdapter, ValidationError

from unique_toolkit._common.execution import (
    failsafe,
)
from unique_toolkit.content.schemas import ContentChunk
from unique_toolkit.language_model.constants import (
    DEFAULT_COMPLETE_TEMPERATURE,
)
from unique_toolkit.language_model.functions import (
    SearchContext,
    _clamp_temperature,
    _to_search_context,
)
from unique_toolkit.language_model.infos import (
    LanguageModelInfo,
    LanguageModelName,
)
from unique_toolkit.language_model.schemas import (
    LanguageModelAssistantMessage,
    LanguageModelMessage,
    LanguageModelMessageOptions,
    LanguageModelMessageRole,
    LanguageModelMessages,
    LanguageModelSystemMessage,
    LanguageModelToolDescription,
    LanguageModelToolMessage,
    LanguageModelUserMessage,
    ResponsesLanguageModelStreamResponse,
)

logger = logging.getLogger(__name__)


def _convert_tools_to_openai(
    tools: Sequence[LanguageModelToolDescription | ToolParam],
) -> list[ToolParam]:
    openai_tools = []
    for tool in tools:
        if isinstance(tool, LanguageModelToolDescription):
            openai_tools.append(tool.to_openai(mode="responses"))
        else:
            openai_tools.append(tool)
    return openai_tools


def _convert_message_to_openai(
    message: LanguageModelMessageOptions,
) -> ResponseInputParam:
    res = []
    match message:
        case LanguageModelAssistantMessage():
            return message.to_openai(mode="responses")  # type: ignore
        case (
            LanguageModelUserMessage()
            | LanguageModelSystemMessage()
            | LanguageModelToolMessage()
        ):
            return [message.to_openai(mode="responses")]
        case _:
            return _convert_message_to_openai(_convert_to_specific_message(message))
    return res


def _convert_to_specific_message(
    message: LanguageModelMessage,
) -> "LanguageModelSystemMessage | LanguageModelUserMessage | LanguageModelAssistantMessage":
    match message.role:
        case LanguageModelMessageRole.SYSTEM:
            return LanguageModelSystemMessage(content=message.content)
        case LanguageModelMessageRole.USER:
            return LanguageModelUserMessage(content=message.content)
        case LanguageModelMessageRole.ASSISTANT:
            return LanguageModelAssistantMessage(content=message.content)
        case LanguageModelMessageRole.TOOL:
            raise ValueError(
                "Cannot convert message with role `tool`. Please use `LanguageModelToolMessage` instead."
            )


def _convert_messages_to_openai(
    messages: Sequence[
        ResponseInputItemParam | LanguageModelMessageOptions | ResponseOutputItem
    ],
) -> ResponseInputParam:
    res = []
    for message in messages:
        if isinstance(message, LanguageModelMessageOptions):
            res.extend(_convert_message_to_openai(message))
        elif isinstance(
            message, dict
        ):  # Openai uses dicts for their input and BaseModel as output
            res.append(message)
        else:
            assert isinstance(message, BaseModel)
            res.append(message.model_dump(exclude_defaults=True))

    return res


class _ResponsesParams(NamedTuple):
    temperature: float
    model_name: str
    search_context: SearchContext | None
    messages: str | ResponseInputParam
    tools: list[ToolParam] | None
    reasoning: Reasoning | None
    text: ResponseTextConfigParam | None


def _prepare_responses_params_util(
    model_name: LanguageModelName | str,
    content_chunks: list[ContentChunk] | None,
    temperature: float,
    tools: Sequence[LanguageModelToolDescription | ToolParam] | None,
    messages: str
    | LanguageModelMessages
    | Sequence[
        ResponseInputItemParam | LanguageModelMessageOptions | ResponseOutputItem
    ],
    reasoning: Reasoning | None,
    text: ResponseTextConfigParam | None,
    other_options: dict | None = None,
) -> _ResponsesParams:
    search_context = (
        _to_search_context(content_chunks) if content_chunks is not None else None
    )

    model = model_name.name if isinstance(model_name, LanguageModelName) else model_name

    tools_res = _convert_tools_to_openai(tools) if tools is not None else None

    if other_options is not None:
        # Key word argument takes precedence
        reasoning = reasoning or _attempt_extract_reasoning_from_options(other_options)
        text = text or _attempt_extract_verbosity_from_options(other_options)

    if isinstance(model_name, LanguageModelName):
        model_info = LanguageModelInfo.from_name(model_name)

        if model_info.temperature_bounds is not None and temperature is not None:
            temperature = _clamp_temperature(temperature, model_info.temperature_bounds)

        if (
            reasoning is None
            and model_info.default_options is not None
            and "reasoning_effort" in model_info.default_options
        ):
            reasoning = Reasoning(effort=model_info.default_options["reasoning_effort"])

        if (
            reasoning is not None
            and tools_res is not None
            and any(tool["type"] == "code_interpreter" for tool in tools_res)
            and "effort" in reasoning
            and reasoning["effort"] == "minimal"
        ):
            logger.warning(
                "Code interpreter cannot be used with `minimal` effort. Switching to `low`."
            )
            reasoning["effort"] = (
                "low"  # Code interpreter cannot be used with minimal effort
            )

    messages_res = None
    if isinstance(messages, LanguageModelMessages):
        messages_res = _convert_messages_to_openai(messages.root)
    elif isinstance(messages, list):
        messages_res = _convert_messages_to_openai(messages)
    else:
        assert isinstance(messages, str)
        messages_res = messages

    return _ResponsesParams(
        temperature, model, search_context, messages_res, tools_res, reasoning, text
    )


@failsafe(
    failure_return_value=None,
    exceptions=(ValidationError,),
    log_exc_info=False,
    logger=logger,
)
def _attempt_extract_reasoning_from_options(
    options: dict[str, Any],
) -> Reasoning | None:
    reasoning: dict[str, Any] | str | None = None

    # Responses API
    if "reasoning" in options:
        reasoning = options["reasoning"]
        # Handle case where reasoning is stored as JSON string (UI limitation)
        if isinstance(reasoning, str):
            try:
                reasoning = json.loads(reasoning)
            except (json.JSONDecodeError, TypeError):
                logger.warning(
                    f"Failed to parse reasoning as JSON string: {reasoning}. "
                    "Continuing with raw value."
                )

    # Completions API
    elif "reasoning_effort" in options:
        reasoning = {"effort": options["reasoning_effort"]}
    if "reasoningEffort" in options:
        reasoning = {"effort": options["reasoningEffort"]}

    if reasoning is not None:
        return TypeAdapter(Reasoning).validate_python(reasoning)

    return None


@failsafe(
    failure_return_value=None,
    exceptions=(ValidationError,),
    log_exc_info=False,
    logger=logger,
)
def _attempt_extract_verbosity_from_options(
    options: dict[str, Any],
) -> ResponseTextConfigParam | None:
    if "verbosity" in options:
        return TypeAdapter(ResponseTextConfigParam).validate_python(
            {"verbosity": options["verbosity"]}
        )

    # Responses API
    if "text" in options:
        text_config: dict[str, Any] | str = options["text"]
        # Handle case where text is stored as JSON string (UI limitation)
        if isinstance(text_config, str):
            try:
                text_config = json.loads(text_config)
                return TypeAdapter(ResponseTextConfigParam).validate_python(text_config)
            except (json.JSONDecodeError, TypeError):
                logger.warning(
                    f"Failed to parse text as JSON string: {text_config}. "
                    "Continuing with raw value."
                )
        if isinstance(text_config, dict):
            return TypeAdapter(ResponseTextConfigParam).validate_python(text_config)

    return None


def _prepare_responses_args(
    company_id: str,
    user_id: str,
    assistant_message_id: str,
    user_message_id: str,
    chat_id: str,
    assistant_id: str,
    params: _ResponsesParams,
    debug_info: dict | None,
    start_text: str | None,
    include: list[ResponseIncludable] | None,
    instructions: str | None,
    max_output_tokens: int | None,
    metadata: Metadata | None,
    parallel_tool_calls: bool | None,
    tool_choice: response_create_params.ToolChoice | None,
    top_p: float | None,
    other_options: dict | None = None,
) -> dict[str, Any]:
    options = {}

    options["company_id"] = company_id
    options["user_id"] = user_id

    options["model"] = params.model_name

    if params.search_context is not None:
        options["searchContext"] = params.search_context

    options["chatId"] = chat_id
    options["assistantId"] = assistant_id
    options["assistantMessageId"] = assistant_message_id
    options["userMessageId"] = user_message_id

    if debug_info is not None:
        options["debugInfo"] = debug_info
    if start_text is not None:
        options["startText"] = start_text

    options["input"] = params.messages

    openai_options: unique_sdk.Integrated.CreateStreamResponsesOpenaiParams = {}

    explicit_options = {
        "temperature": params.temperature,
        "reasoning": params.reasoning,
        "text": params.text,
        "include": include,
        "instructions": instructions,
        "max_output_tokens": max_output_tokens,
        "metadata": metadata,
        "parallel_tool_calls": parallel_tool_calls,
        "tool_choice": tool_choice,
        "tools": params.tools,
        "top_p": top_p,
    }

    openai_options.update({k: v for k, v in explicit_options.items() if v is not None})  # type: ignore[arg-type]

    # allow any other openai.resources.responses.Response.create options
    if other_options is not None:
        for k, v in other_options.items():
            openai_options.setdefault(k, v)  # type: ignore

    options["options"] = openai_options

    return options


def stream_responses_with_references(
    *,
    company_id: str,
    user_id: str,
    assistant_message_id: str,
    user_message_id: str,
    chat_id: str,
    assistant_id: str,
    model_name: LanguageModelName | str,
    messages: str
    | LanguageModelMessages
    | Sequence[
        ResponseInputItemParam | LanguageModelMessageOptions | ResponseOutputItem
    ],
    content_chunks: list[ContentChunk] | None = None,
    tools: Sequence[LanguageModelToolDescription | ToolParam] | None = None,
    temperature: float = DEFAULT_COMPLETE_TEMPERATURE,
    debug_info: dict | None = None,
    start_text: str | None = None,
    include: list[ResponseIncludable] | None = None,
    instructions: str | None = None,
    max_output_tokens: int | None = None,
    metadata: Metadata | None = None,
    parallel_tool_calls: bool | None = None,
    text: ResponseTextConfigParam | None = None,
    tool_choice: response_create_params.ToolChoice | None = None,
    top_p: float | None = None,
    reasoning: Reasoning | None = None,
    other_options: dict | None = None,
) -> ResponsesLanguageModelStreamResponse:
    responses_params = _prepare_responses_params_util(
        model_name=model_name,
        content_chunks=content_chunks,
        temperature=temperature,
        tools=tools,
        messages=messages,
        reasoning=reasoning,
        text=text,
        other_options=other_options,
    )

    responses_args = _prepare_responses_args(
        company_id=company_id,
        user_id=user_id,
        assistant_message_id=assistant_message_id,
        user_message_id=user_message_id,
        chat_id=chat_id,
        assistant_id=assistant_id,
        params=responses_params,
        debug_info=debug_info,
        start_text=start_text,
        include=include,
        instructions=instructions,
        max_output_tokens=max_output_tokens,
        metadata=metadata,
        parallel_tool_calls=parallel_tool_calls,
        tool_choice=tool_choice,
        top_p=top_p,
        other_options=other_options,
    )

    return ResponsesLanguageModelStreamResponse.model_validate(
        unique_sdk.Integrated.responses_stream(
            **responses_args,
        )
    )


async def stream_responses_with_references_async(
    *,
    company_id: str,
    user_id: str,
    assistant_message_id: str,
    user_message_id: str,
    chat_id: str,
    assistant_id: str,
    model_name: LanguageModelName | str,
    messages: str
    | LanguageModelMessages
    | Sequence[
        ResponseInputItemParam | LanguageModelMessageOptions | ResponseOutputItem
    ],
    content_chunks: list[ContentChunk] | None = None,
    tools: Sequence[LanguageModelToolDescription | ToolParam] | None = None,
    temperature: float = DEFAULT_COMPLETE_TEMPERATURE,
    debug_info: dict | None = None,
    start_text: str | None = None,
    include: list[ResponseIncludable] | None = None,
    instructions: str | None = None,
    max_output_tokens: int | None = None,
    metadata: Metadata | None = None,
    parallel_tool_calls: bool | None = None,
    text: ResponseTextConfigParam | None = None,
    tool_choice: response_create_params.ToolChoice | None = None,
    top_p: float | None = None,
    reasoning: Reasoning | None = None,
    other_options: dict | None = None,
) -> ResponsesLanguageModelStreamResponse:
    responses_params = _prepare_responses_params_util(
        model_name=model_name,
        content_chunks=content_chunks,
        temperature=temperature,
        tools=tools,
        messages=messages,
        reasoning=reasoning,
        text=text,
        other_options=other_options,
    )

    responses_args = _prepare_responses_args(
        company_id=company_id,
        user_id=user_id,
        assistant_message_id=assistant_message_id,
        user_message_id=user_message_id,
        chat_id=chat_id,
        assistant_id=assistant_id,
        params=responses_params,
        debug_info=debug_info,
        start_text=start_text,
        include=include,
        instructions=instructions,
        max_output_tokens=max_output_tokens,
        metadata=metadata,
        parallel_tool_calls=parallel_tool_calls,
        tool_choice=tool_choice,
        top_p=top_p,
        other_options=other_options,
    )

    return ResponsesLanguageModelStreamResponse.model_validate(
        await unique_sdk.Integrated.responses_stream_async(
            **responses_args,
        )
    )



from typing import Annotated, Any

from pydantic import BaseModel, Field

from unique_toolkit._common.config_checker import register_config
from unique_toolkit._common.validators import LMI, get_LMI_default_field
from unique_toolkit.agentic.evaluation.context_relevancy.schema import (
    StructuredOutputConfig,
)
from unique_toolkit.agentic.tools.config import get_configuration_dict
from unique_toolkit.language_model.default_language_model import DEFAULT_GPT_4o


@register_config()
class ChunkRelevancySortConfig(BaseModel):
    model_config = get_configuration_dict()
    enabled: bool = Field(
        default=False,
        description="Whether to enable the chunk relevancy sort.",
    )
    relevancy_levels_to_consider: list[str] = Field(
        default=["high", "medium", "low"],
        description="The relevancy levels to consider.",
    )
    relevancy_level_order: dict[str, int] = Field(
        default={"high": 0, "medium": 1, "low": 2},
        description="The relevancy level order.",
    )
    language_model: LMI = get_LMI_default_field(
        DEFAULT_GPT_4o,
        description="The language model to use for the chunk relevancy sort.",
    )
    fallback_language_model: LMI = get_LMI_default_field(
        DEFAULT_GPT_4o,
        description="The language model to use as a fallback.",
    )
    additional_llm_options: dict[str, Any] = Field(
        default={},
        description="Additional options to pass to the language model.",
    )
    structured_output_config: StructuredOutputConfig = Field(
        default_factory=StructuredOutputConfig,
        description="The configuration for the structured output.",
    )
    max_tasks: (
        Annotated[int, Field(title="Limited")]
        | Annotated[None, Field(title="Unlimited")]
    ) = Field(
        default=50,
        description="The maximum number of tasks to run in parallel.",
    )



"""Services for the write-up agent pipeline."""

from unique_toolkit._common.experimental.write_up_agent.services.dataframe_handler import (
    DataFrameHandler,
)
from unique_toolkit._common.experimental.write_up_agent.services.template_handler import (
    TemplateHandler,
)

__all__ = [
    "DataFrameHandler",
    "TemplateHandler",
]



import json
import logging
import os
from typing import Callable, TypeVar

import unique_sdk
from pydantic import ValidationError

from unique_toolkit.app.schemas import Event

logger = logging.getLogger(f"toolkit.{__name__}")


class WebhookVerificationError(Exception):
    """Custom exception for webhook verification errors."""

    pass


T = TypeVar("T")


def verify_signature_and_construct_event(
    headers: dict[str, str],
    payload: bytes,
    endpoint_secret: str,
    logger: logging.Logger = logger,
    event_constructor: Callable[..., T] = Event,
) -> T:
    """
    Verify the signature of a webhook and construct an event object.

    Args:
        headers (Dict[str, str]): The headers of the webhook request.
        payload (bytes): The raw payload of the webhook request.
        endpoint_secret (str): The secret used to verify the webhook signature.
        logger (logging.Logger): A logger instance for logging messages.
        event_constructor (Callable[..., T]): A callable that constructs an event object.
    Returns:
        T: The constructed event object.

    Raises:
        WebhookVerificationError: If there's an error during verification or event construction.
    """

    sig_header = headers.get("X-Unique-Signature")
    timestamp = headers.get("X-Unique-Created-At")

    if not sig_header or not timestamp:
        logger.error("⚠️  Webhook signature or timestamp headers missing.")
        raise WebhookVerificationError("Signature or timestamp headers missing")

    try:
        event = unique_sdk.Webhook.construct_event(
            payload,
            sig_header,
            timestamp,
            endpoint_secret,
        )
        logger.info("✅  Webhook signature verification successful.")
        return event_constructor(**event)
    except unique_sdk.SignatureVerificationError as e:
        logger.error("⚠️  Webhook signature verification failed. " + str(e))
        raise WebhookVerificationError(f"Signature verification failed: {str(e)}")


def verify_request_and_construct_event(
    assistant_name: str,
    payload: bytes,
    headers: dict[str, str],
    event_constructor: Callable[..., Event] = Event,
) -> tuple[str, int] | tuple[Event, int]:
    """Check the payload, authenticate and genenrate the event if the payload is correct"""
    logger.info(f"{assistant_name} - received request")

    try:
        payload_decoded = json.loads(payload)
    except json.decoder.JSONDecodeError as e:
        logger.error(f"Error decoding payload: {e}", exc_info=True)
        return "Invalid payload", 400

    endpoint_secret = os.environ.get("ENDPOINT_SECRET", None)
    if endpoint_secret:
        response = verify_signature_and_construct_event(
            headers=headers,  # type: ignore
            payload=payload,
            endpoint_secret=endpoint_secret,
            logger=logger,
            event_constructor=event_constructor,
        )
        if isinstance(response, tuple):
            return response  # Error response
        event = response  # This is an event since it is not a tuple
    else:
        try:
            event = event_constructor(**payload_decoded)
        except ValidationError as e:
            # pydantic errors https://docs.pydantic.dev/2.10/errors/errors/
            logger.error(f"Validation error with model: {e.json()}", exc_info=True)
            raise ValidationError(e)
        except ValueError as e:
            logger.error(f"Error deserializing event: {e}", exc_info=True)
            return "Invalid event", 400

    if not event.payload.name == assistant_name:
        logger.error(
            f"{assistant_name}: Incorrect assistant: {event.payload.name}: Expected {assistant_name}"
        )
        return f"Not {assistant_name} event", 400

    logger.info(f"{assistant_name} - received event")
    return event, 200



from typing import Any

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from unique_toolkit._common.config_checker import register_config


class FeatureFlag:
    """A feature flag that can be enabled globally or for specific company IDs.

    Examples:
        >>> flag = FeatureFlag(True)
        >>> flag.is_enabled("any_company")
        True

        >>> flag = FeatureFlag(["company1", "company2"])
        >>> flag.is_enabled("company1")
        True
        >>> flag.is_enabled("company3")
        False
    """

    def __init__(self, value: list[str] | bool):
        self.value = value

    def is_enabled(self, company_id: str | None = None) -> bool:
        """Check if the feature is enabled for the given company.

        Args:
            company_id: The company ID to check. Required if flag is a list.

        Returns:
            True if the feature is enabled, False otherwise.
        """

        if isinstance(self.value, bool):
            return self.value

        if isinstance(self.value, list):
            return company_id in self.value if company_id else False

        return False

    def __repr__(self) -> str:
        return f"FeatureFlag({self.value})"


@register_config()
class FeatureFlags(BaseSettings):
    """Feature flags loaded from environment variables.

    Environment variables are automatically loaded based on field names.
    For example, `enable_new_answers_ui_un_14411` will be loaded from
    `FEATURE_FLAG_ENABLE_NEW_ANSWERS_UI_UN_14411` (with the env_prefix).

    Usage:
        >>> flags = FeatureFlags()
        >>> # Check if feature is enabled for a company
        >>> flags.enable_new_answers_ui_un_14411.is_enabled("company123")
        >>> # Check global enablement
        >>> bool(flags.enable_elicitation_un_15809)
    """

    enable_new_answers_ui_un_14411: FeatureFlag = Field(
        default=FeatureFlag([]),
        description="Enable new answers UI (UN-14411)",
    )

    enable_elicitation_un_15809: FeatureFlag = Field(
        default=FeatureFlag(False),
        description="Enable elicitation (UN-15809)",
    )

    enable_html_rendering_un_15131: FeatureFlag = Field(
        default=FeatureFlag(False),
        description="Enable HTML rendering for code interpreter files (UN-15131). Can be 'true' or comma-separated company IDs.",
    )

    model_config = SettingsConfigDict(
        extra="ignore",
        env_prefix="FEATURE_FLAG_",
        case_sensitive=False,
        env_file=".env",
        env_file_encoding="utf-8",
    )

    @field_validator("*", mode="before")
    @classmethod
    def parse_feature_flag(cls, v: Any) -> FeatureFlag:
        """Parse all feature flag fields from environment variables.

        Args:
            v: Can be a string ("true", "false", or comma-separated IDs),
               a boolean, a list of IDs, or already a FeatureFlag instance.

        Returns:
            A FeatureFlag instance.
        """
        if isinstance(v, FeatureFlag):
            return v

        if isinstance(v, str):
            v_lower = v.lower().strip()
            if v_lower in ("true", "1", "yes"):
                return FeatureFlag(True)
            elif v_lower in ("false", "0", "no", ""):
                return FeatureFlag(False)
            else:
                # Comma-separated company IDs
                return FeatureFlag([id.strip() for id in v.split(",") if id.strip()])

        if isinstance(v, bool):
            return FeatureFlag(v)

        if isinstance(v, list):
            return FeatureFlag(v)

        # Default to disabled (this handles default factory functions too)
        return FeatureFlag(False)


# Initialize once at module load - import this where needed
feature_flags = FeatureFlags()



import asyncio
import logging
import threading
import time
from typing import (
    Awaitable,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

T = TypeVar("T")
Result = Union[T, BaseException]


async def run_async_tasks_parallel(
    tasks: Sequence[Awaitable[T]],
    max_tasks: Optional[int] = None,
    logger: logging.Logger = logging.getLogger(__name__),
) -> list[Result]:
    """
    Executes the a set of given async tasks and returns the results.

    Args:
    tasks (list[Awaitable[T]]): list of async callables to execute in parallel.
    max_tasks (int): Maximum number of tasks for the asyncio Semaphore.

    Returns:
    list[Result]: list of results from the executed tasks.
    """

    max_tasks = max_tasks or len(tasks)

    async def logging_wrapper(task: Awaitable[T], task_id: int) -> Result:
        thread = threading.current_thread()
        start_time = time.time()

        logger.info(
            f"Thread {thread.name} (ID: {thread.ident}) starting task {task_id}"
        )

        try:
            result = await task
            return result
        except Exception as e:
            logger.error(
                f"Thread {thread.name} (ID: {thread.ident}) - Task {task_id} failed with error: {e}"
            )
            return e
        finally:
            end_time = time.time()
            duration = end_time - start_time
            logger.debug(
                f"Thread {thread.name} (ID: {thread.ident}) - Task {task_id} finished in {duration:.2f} seconds"
            )

    sem = asyncio.Semaphore(max_tasks)

    async def sem_task(task: Awaitable[T], task_id: int) -> Result:
        async with sem:
            return await logging_wrapper(task, task_id)

    wrapped_tasks: list[Awaitable[Result]] = [
        sem_task(task, i) for i, task in enumerate(tasks)
    ]

    results: list[Result] = await asyncio.gather(*wrapped_tasks, return_exceptions=True)

    return results



from typing import Any, Protocol, Required, Sequence, Unpack

from openai.types.chat import ChatCompletionNamedToolChoiceParam
from openai.types.responses import (
    ResponseIncludable,
    ResponseInputItemParam,
    ResponseOutputItem,
    ResponseTextConfigParam,
    ToolParam,
    response_create_params,
)
from openai.types.shared_params import Metadata, Reasoning
from typing_extensions import TypedDict

from unique_toolkit import LanguageModelToolDescription
from unique_toolkit.chat.functions import LanguageModelStreamResponse
from unique_toolkit.chat.service import LanguageModelMessages
from unique_toolkit.content import ContentChunk
from unique_toolkit.language_model.infos import LanguageModelInfo
from unique_toolkit.language_model.schemas import (
    LanguageModelMessageOptions,
    ResponsesLanguageModelStreamResponse,
)
from unique_toolkit.protocols.support import (
    ResponsesSupportCompleteWithReferences,
    SupportCompleteWithReferences,
)


class _LoopIterationRunnerKwargs(TypedDict, total=False):
    iteration_index: Required[int]
    streaming_handler: Required[SupportCompleteWithReferences]
    messages: Required[LanguageModelMessages]
    model: Required[LanguageModelInfo]
    tools: list[LanguageModelToolDescription]
    content_chunks: list[ContentChunk]
    start_text: str
    debug_info: dict[str, Any]
    temperature: float
    tool_choices: list[ChatCompletionNamedToolChoiceParam]
    other_options: dict[str, Any]


class LoopIterationRunner(Protocol):
    """
    A loop iteration runner is responsible for running a single iteration of the loop, and returning the stream response for that iteration.
    """

    async def __call__(
        self,
        **kwargs: Unpack[_LoopIterationRunnerKwargs],
    ) -> LanguageModelStreamResponse: ...


class _ResponsesLoopIterationRunnerKwargs(TypedDict, total=False):
    iteration_index: Required[int]
    streaming_handler: Required[ResponsesSupportCompleteWithReferences]
    messages: Required[
        str
        | LanguageModelMessages
        | Sequence[
            ResponseInputItemParam | LanguageModelMessageOptions | ResponseOutputItem
        ]
    ]
    model: Required[LanguageModelInfo]
    tools: list[LanguageModelToolDescription | ToolParam]
    content_chunks: list[ContentChunk]
    start_text: str
    debug_info: dict[str, Any]
    temperature: float
    include: list[ResponseIncludable]
    instructions: str
    max_output_tokens: int
    metadata: Metadata
    parallel_tool_calls: bool
    text: ResponseTextConfigParam
    tool_choices: list[response_create_params.ToolChoice]
    top_p: float
    reasoning: Reasoning
    other_options: dict[str, Any]


class ResponsesLoopIterationRunner(Protocol):
    async def __call__(
        self,
        **kwargs: Unpack[_ResponsesLoopIterationRunnerKwargs],
    ) -> ResponsesLanguageModelStreamResponse: ...
